import gc
import numpy as np
from scipy import linalg
from scipy import sparse
from scipy.sparse import csr_matrix, isspmatrix, block_diag, lil_matrix, hstack, vstack
from scipy.sparse.linalg import inv


def pre_covar_inv(covar_dim, var_com):
    # inversion of covariance matrix
    end = 0
    covar_inv = []  # covariance matrix list
    if var_com[-1] <= 0:
        return None
    for i in range(len(covar_dim)):
        start = end
        ind_sel = np.tril_indices(covar_dim[i])
        end = len(ind_sel[0]) + start
        covar = np.zeros((covar_dim[i], covar_dim[i]))
        covar[ind_sel] = var_com[start:end]
        covar = np.add(covar, np.tril(covar, -1).T)
        try:
            linalg.cholesky(covar)
        except Exception as e:
            # print(e)
            return None
        covar = linalg.inv(covar)
        covar_inv.append(covar)
    
    return covar_inv


def pre_coef_mat(y, xmat, eff_ind, zmat_con, covar_dim, var_com, covar_inv, coef_mat_pure, rmat_inv_lst, rhs_pure):
    # Coefficient Matrix
    coef_mat = (coef_mat_pure.multiply(1.0 / var_com[-1])).toarray()
    for i in range(len(covar_dim)):
        if isspmatrix(rmat_inv_lst[i]):
            temp = sparse.kron(covar_inv[i], rmat_inv_lst[i])
            temp = temp.toarray()
        else:
            temp = linalg.kron(covar_inv[i], rmat_inv_lst[i])
        
        coef_mat[eff_ind[i + 1][0]:eff_ind[i + 1][-1], eff_ind[i + 1][0]:eff_ind[i + 1][-1]] = \
            np.add(coef_mat[eff_ind[i + 1][0]:eff_ind[i + 1][-1], eff_ind[i + 1][0]:eff_ind[i + 1][-1]], temp)
    
    # right hand
    rhs_mat = np.divide(rhs_pure, var_com[-1])
    
    # all effects
    coef_mat = linalg.inv(coef_mat)
    eff = np.dot(coef_mat, rhs_mat)
    
    # errors
    e = y - xmat.dot(eff[:eff_ind[0][1], :]) - zmat_con.dot(
        eff[eff_ind[0][1]:, :])
    
    return coef_mat, eff, e




### gma.unbalance_varcom
import numpy as np
import pandas as pd
from gma import cal_kin
from gma import unbalance_varcom
bed_file = './tests/plink'
kin_lst = cal_kin(bed_file)
data_file = './tests/phe.unbalance.txt'
id = 'ID'
tpoint = 'weak'
trait = 'trait'
kin_inv_file = './tests/plink.giv'
tfix = 'Sex'
prefix_outfile = './tests/gma_unbalance_varcom'
res_var = unbalance_varcom(data_file, id, tpoint, trait, kin_inv_file, tfix=tfix, prefix_outfile=prefix_outfile)
print res_var


from gma import rrm_sparse_varcom
res_var = rrm_sparse_varcom(data_file, id, tpoint, trait, kin_inv_file, tfix=tfix, prefix_outfile='./tests/gma_rrm_sparse_varcom')




### gma.unbalance_longwas_fixreg
import numpy as np
import pandas as pd
from gma import cal_kin
from gma import unbalance_varcom
from gma import unbalance_longwas_fixreg
bed_file = './tests/plink'
kin_lst = cal_kin(bed_file)
data_file = './tests/phe.unbalance.txt'
id = 'ID'
tpoint = 'weak'
trait = 'trait'
kin_inv_file = './tests/plink.giv'
tfix = 'Sex'
prefix_outfile = './tests/gma_unbalance_varcom'
res_var = unbalance_varcom(data_file, id, tpoint, trait, kin_inv_file, tfix=tfix, prefix_outfile=prefix_outfile)
print res_var

kin_file = './tests/plink.grm'
var_com = pd.read_csv("./tests/gma_unbalance_varcom.var", sep='\s+', header=0)
prefix_outfile = './tests/gma_unbalance_longwas_fixreg'
res_lst = unbalance_longwas_fixreg(data_file, id, tpoint, trait, bed_file, kin_file, kin_inv_file, var_com, tfix=tfix,
                             prefix_outfile=prefix_outfile)

### gma.unbalance_longwas_lt
import numpy as np
import pandas as pd
from gma import cal_kin
from gma import unbalance_varcom
from gma import unbalance_longwas_lt
bed_file = './tests/plink'
kin_lst = cal_kin(bed_file)
data_file = './tests/phe.unbalance.txt'
id = 'ID'
tpoint = 'weak'
trait = 'trait'
kin_inv_file = './tests/plink.giv'
tfix = 'Sex'
prefix_outfile = './tests/gma_unbalance_varcom'
res_var = unbalance_varcom(data_file, id, tpoint, trait, kin_inv_file, tfix=tfix, prefix_outfile=prefix_outfile)
print res_var

kin_file = './tests/plink.grm'
var_com = pd.read_csv("./tests/gma_unbalance_varcom.var", sep='\s+', header=0)
prefix_outfile = './tests/gma_unbalance_longwas_lt'
res_lst = unbalance_longwas_lt(data_file, id, tpoint, trait, bed_file, kin_file, kin_inv_file, var_com, tfix=tfix,
                             prefix_outfile=prefix_outfile)




import numpy as np
import pandas as pd
from _cread_plink_bed import ffi, lib


class Bed():
    """
    描述：
    读入plink的bed文件
    """

    def __init__(self, bed_file):
        self.bed_file = bed_file
        self.snp_info = pd.read_csv(bed_file + '.bim', header=None, sep='\s+')
        self.num_snp = self.snp_info.shape[0]
        self.id_info = pd.read_csv(bed_file + '.fam', header=None, sep='\s+')
        self.num_id = self.id_info.shape[0]

    def read(self):
        pbed_file = ffi.new("char[]", self.bed_file.encode('ascii'))
        pnum_id = ffi.cast("long long", self.num_id)
        pnum_snp = ffi.cast("long long", self.num_snp)
        marker_mat = np.ones(self.num_id * self.num_snp, dtype=np.float64)
        pmarker_mat = ffi.cast("double *", ffi.from_buffer(marker_mat))
        lib.read_plink_bed(pbed_file, pnum_id, pnum_snp, pmarker_mat)
        marker_mat[np.abs(marker_mat - 1.0/3) < 0.0001] = np.nan
        marker_mat.shape = self.num_snp, self.num_id
        return marker_mat.T



import numpy as np
import pandas as pd
import time
from glta.process_plink.process_plink import read_plink, impute_geno


def cal_agmat(bed_file, inv=True, small_val=0.001):
    print("{:#^80}".format("Read the SNP data"))
    snp_mat = read_plink(bed_file)
    if np.any(np.isnan(snp_mat)):
        print('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    print("There are {:d} individuals and {:d} SNPs.".format(snp_mat.shape[0], snp_mat.shape[1]))
    freq = np.sum(snp_mat, axis=0) / (2 * snp_mat.shape[0])
    freq.shape = (1, snp_mat.shape[1])
    scale = 2 * freq * (1 - freq)
    scale = np.sum(scale)
    snp_mat = snp_mat - 2 * freq
    
    print("{:#^80}".format("Calculate the additive genomic relationship matrix"))
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    kin = np.dot(snp_mat, snp_mat.T) / scale
    kin_diag = np.diag(kin)
    kin_diag = kin_diag + kin_diag * small_val
    np.fill_diagonal(kin, kin_diag)
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    print("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    
    print("{:#^80}".format("Output"))
    ind = np.tril_indices_from(kin)
    val = kin[ind]
    fam_info = pd.read_csv(bed_file + '.fam', sep='\s+', header=None)
    id1 = np.array(fam_info.iloc[ind[0], 1])
    id2 = np.array(fam_info.iloc[ind[1], 1])
    data_df = {'id1': id1, 'id2': id2, 'val': val}
    data_df = pd.DataFrame(data_df, columns=['id1', 'id2', 'val'])
    kin_file = bed_file + '.grm'
    print("The output file is", kin_file)
    data_df.to_csv(kin_file, sep=' ', header=False, index=False)
    
    kin_inv = None
    if inv:
        print("{:#^80}".format("Calculate the inversion of kinship"))
        clock_t0 = time.perf_counter()
        cpu_t0 = time.process_time()
        kin_inv = np.linalg.inv(kin)
        clock_t1 = time.perf_counter()
        cpu_t1 = time.process_time()
        print("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0,
                                                                                   cpu_t1 - cpu_t0))
        
        print("{:#^80}".format("Output the inversion"))
        val = kin_inv[ind]
        data_df = {'id1': id1, 'id2': id2, 'val': val}
        data_df = pd.DataFrame(data_df, columns=['id1', 'id2', 'val'])
        kin_inv_file = bed_file + '.giv'
        print("The output file is", kin_inv_file)
        data_df.to_csv(kin_inv_file, sep=' ', header=False, index=False)
    
    return kin, kin_inv


def cal_ginbreedcoef(bed_file):
    snp_mat = read_plink(bed_file)
    if np.any(np.isnan(snp_mat)):
        print('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    print("There are {:d} individuals and {:d} SNPs.".format(snp_mat.shape[0], snp_mat.shape[1]))
    homo_f = 1 - np.sum(np.abs(snp_mat - 1.0) < 0.01, axis=1) / snp_mat.shape[1]
    freq = np.sum(snp_mat, axis=0) / (2 * snp_mat.shape[0])
    freq.shape = (1, snp_mat.shape[1])
    scale_vec = 2 * freq * (1 - freq)
    scale = np.sum(scale_vec)
    snp_mat = snp_mat - 2 * freq
    grm_f1 = np.sum(np.multiply(snp_mat, snp_mat), axis=1) / scale - 1.0
    grm_f2 = np.sum(np.multiply(snp_mat, snp_mat) / scale_vec, axis=1) / snp_mat.shape[1] - 1.0
    
    fam_info = pd.read_csv(bed_file + '.fam', sep='\s+', header=None)
    id = np.array(fam_info.iloc[:, 1])
    data_df = {'id': id, 'homo_F': homo_f, 'grm_F1': grm_f1, 'grm_F2': grm_f2}
    data_df = pd.DataFrame(data_df, columns=['id', 'homo_F', 'grm_F1', 'grm_F2'])
    out_file = bed_file + '.ginbreedcoef'
    data_df.to_csv(out_file, sep=' ', header=True, index=False)
    